public class LinearHoughTransform implements PixelBufferFilter {
	// TODO: implement as a parameter
	private final static boolean AXIS = false;		// true=X, false=Y
	
	// ### FILTER ###
	
	private int width = -1, height = -1;
	
	/* recycled arrays */
	private int[] inGrayPixels = null;
	private int[][] xPlaneSums = null;
	private int[][] yPlaneSums = null;
	
	public void filter(int width, int height, int[] pixels) {
		if ((width != this.width) || (height != this.height)) {
			this.width = width;
			this.height = height;
			
			if (AXIS == true) {
				this.xPlaneSums = new int[height*2+1][width*3];
			} else {
				this.yPlaneSums = new int[width*2+1][height*3];
			}
		}
		
		/* Obtain grayscale image from input image */
		this.inGrayPixels = Pixels.ARGBPixelsToGrayPixels(pixels, this.inGrayPixels);
		
		long startTime = System.currentTimeMillis();		// DEBUG: remove
		
		if (AXIS == true) {
			/* Calculate xPlane sums */
			int minIntercept = -width+1;
			int limitIntercept = width*2;
			for (int slopeDenom = -height, i=0; slopeDenom <= height; slopeDenom++, i++) {
				int[] sumsForCurSlopeDenom = xPlaneSums[i];
				for (int intercept=minIntercept, j=0; intercept<limitIntercept; intercept++, j++) {
					sumsForCurSlopeDenom[j] = 
						getSumOfGraysAlongLine(
							true,		/* x axis */
							slopeDenom,
							intercept);
				}
			}
		} else {
			/* Calculate yPlane sums */
			int minIntercept = -height+1;
			int limitIntercept = height*2;
			for (int slopeDenom = -width, i=0; slopeDenom <= width; slopeDenom++, i++) {
				int[] sumsForCurSlopeDenom = yPlaneSums[i];
				for (int intercept=minIntercept, j=0; intercept<limitIntercept; intercept++, j++) {
					sumsForCurSlopeDenom[j] = 
						getSumOfGraysAlongLine(
							false,		/* y axis */
							slopeDenom,
							intercept);
				}
			}
		}
		
		long endTime = System.currentTimeMillis();			// DEBUG: remove
		System.out.println("PROF: " + (endTime-startTime));
		
		if (AXIS == true) {
			throw new IllegalStateException("not yet implemented");	// FIXME
		} else {
			/*
			 * Normalize the Hough transform output
			 */
			{
				/* Determine maximum sum */
				int maxSum = 0;
				{
					int minIntercept = -height+1;
					int limitIntercept = height*2;
					for (int slopeDenom = -width, i=0; slopeDenom <= width; slopeDenom++, i++) {
						int[] sumsForCurSlopeDenom = yPlaneSums[i];
						for (int intercept=minIntercept, j=0; intercept<limitIntercept; intercept++, j++) {
							int curSum = sumsForCurSlopeDenom[j];
							if (curSum > maxSum) {
								maxSum = curSum;
							}
						}
					}
				}
				
				/* Normalize sums to fall in the range [0,255] */
				{
					int minIntercept = -height+1;
					int limitIntercept = height*2;
					for (int slopeDenom = -width, i=0; slopeDenom <= width; slopeDenom++, i++) {
						int[] sumsForCurSlopeDenom = yPlaneSums[i];
						for (int intercept=minIntercept, j=0; intercept<limitIntercept; intercept++, j++) {
							sumsForCurSlopeDenom[j] = sumsForCurSlopeDenom[j]*255/maxSum;
						}
					}
				}
			}
			
			
			/*
			 * Output: Scaled version of Hough transform output for the Y plane
			 * 
			 * (width*2 + 1) x (height*3) --> (width) x (height)
			 */
			for (int x=0, i=0, colOff=0; x<width; x++, i+=2, colOff++) {
				int[] sumsForCurSlopeDenom = yPlaneSums[i];
				for (int y=0, j=0, off=colOff; y<height; y++, j+=3, off+=width) {
					pixels[off] = Pixels.grayToARGB(sumsForCurSlopeDenom[j]);
				}
			}
		}
	}
	
	/**
	 * @return		the sum of the gray-levels along the line
	 * 				with slope <tt>1/slopeDenom</tt> and intercept <tt>intercept</tt>
	 * 				on the <tt>axis</tt> (true=X, false=Y) axis
	 */
	// NOTE: this code derived from the LineGenerator filter
	public int getSumOfGraysAlongLine(boolean axis, int slopeDenom, int intercept) {
		int sum = 0;
		
		if (axis == false) {
			/* Y Axis */
			if (slopeDenom > 0) {
				// Optimization: Don't consider lines that are known to be totally offscreen
				if (intercept >= height) {
					return 0;
				}
				
				int x, y;
				if (intercept >= 0) {
					x = 0;
					y = intercept;
				} else {
					x = (0-intercept) * slopeDenom;
					y = 0;
				}
				for (int off=y*width+x;
					 y<height && x<width;
					 y++, off+=width)
				{
					for (int i=slopeDenom;
						 i>0 && x<width;
						 i--, x++, off++)
					{
						sum += inGrayPixels[off];
					}
				}
			} else if (slopeDenom < 0) {
				slopeDenom = -slopeDenom;
				
				// Optimization: Don't consider lines that are known to be totally offscreen
				if (intercept < 0) {
					return 0;
				}
				
				int x, y;
				if (intercept < height) {
					x = 0;
					y = intercept;
				} else {
					x = (intercept-height+1) * slopeDenom;
					y = height-1;
				}
				for (int off=y*width+x;
					 y>=0 && x<width;
					 y--, off-=width)
				{
					for (int i=slopeDenom;
						 i>0 && x<width;
						 i--, x++, off++)
					{
						sum += inGrayPixels[off];
					}
				}
			} else if (slopeDenom == 0) {
				// can't draw a slope-intercept line with a slope of infinity
			}
		} else {
			// FIXME: propagate changes from the "Y Axis" section to this section
			/* X Axis */
			if (slopeDenom > 0) {
				int y = 0;
				int slopeDenom_times_width = slopeDenom*width;
				for (int x=intercept, off=y*width+x;
					 x<width && y<height;
					 x++, off++)
				{
					if (x < 0) {
						/* x is out of bounds (too small) */
						y += slopeDenom;
						off += slopeDenom_times_width;
					} else {
						/* x is in bounds */
						for (int i=slopeDenom;
							 i>0 && y<height;
							 i--, y++, off+=width)
						{
							sum += inGrayPixels[off];
						}
					}
				}
			} else if (slopeDenom < 0) {
				slopeDenom = -slopeDenom;
				
				int y = 0;
				int slopeDenom_times_width = slopeDenom*width;
				for (int x=intercept, off=y*width+x;
					 x>=0 && y<height;
					 x--, off--)
				{
					if (x >= width) {
						/* x is out of bounds (too big) */
						y += slopeDenom;
						off += slopeDenom_times_width;
					} else {
						/* x is in bounds */
						for (int i=slopeDenom;
							 i>0 && y<height;
							 i--, y++, off+=width)
						{
							sum += inGrayPixels[off];
						}
					}
				}
			} else if (slopeDenom == 0) {
				// can't draw a slope-intercept line with a slope of infinity
			}
		}
		
		return sum;
	}
	
	public String toString() {
		return "Linear Hough Transform";
	}
}